
#include "ContextFreeGrammar.h"
using namespace std;




ContextFreeGrammars::Symbol::Symbol(string Name)
{
    this->Name = Name;
}

ContextFreeGrammars::Symbol::~Symbol()
{
}




ContextFreeGrammars::Terminal::Terminal(string Name)
    : Symbol(Name)
{
}

ContextFreeGrammars::Terminal::~Terminal()
{
}




ContextFreeGrammars::Production::Production()
{
}

ContextFreeGrammars::Production::~Production()
{
}




ContextFreeGrammars::NonTerminal::NonTerminal(string Name)
    : Symbol(Name)
{
}

ContextFreeGrammars::NonTerminal::~NonTerminal()
{
    for(vector<Production*>::iterator i = productions.begin(); i != productions.end(); i++)
        delete *i;
    productions.clear();
}

void ContextFreeGrammars::NonTerminal::AddOrDeleteProduction(Production* production)
{
    // test if we already have this production
    bool isNew = true;
    for(vector<Production*>::const_iterator p = productions.begin(); isNew && p != productions.end(); p++)
    {
        bool isEqual = production->produced.size() == (*p)->produced.size();
        int q = 0;
        for(vector<Symbol*>::const_iterator ps = (*p)->produced.begin(), ns = production->produced.begin(); isEqual && ps != (*p)->produced.end(); ps++, ns++)
            if(*ps != *ns)
                isEqual = false;

        if(isEqual)
        {
            delete production;
            return;
        }
    }

    productions.push_back(production);
}



ContextFreeGrammars::ContextFreeGrammar::ContextFreeGrammar()
{
}

ContextFreeGrammars::ContextFreeGrammar::ContextFreeGrammar(const ContextFreeGrammar& cfg)
{
    operator=(cfg);
}

ContextFreeGrammars::ContextFreeGrammar::~ContextFreeGrammar()
{
}

void ContextFreeGrammars::ContextFreeGrammar::Clear()
{
    for(vector<Terminal*>::iterator i = terminals.begin(); i != terminals.end(); i++)
        delete *i;
    terminals.clear();
    for(vector<NonTerminal*>::iterator i = nonterminals.begin(); i != nonterminals.end(); i++)
        delete *i;
    nonterminals.clear();

    symbols.clear();
}

ContextFreeGrammars::ContextFreeGrammar& ContextFreeGrammars::ContextFreeGrammar::operator=(const ContextFreeGrammar& cfg)
{
    Clear();

    // copy terminals
    for(vector<Terminal*>::const_iterator i = cfg.terminals.begin(); i != cfg.terminals.end(); i++)
        AddTerminal((*i)->Name);

    // copy nonterminals
    for(vector<NonTerminal*>::const_iterator i = cfg.nonterminals.begin(); i != cfg.nonterminals.end(); i++)
        AddNonTerminal((*i)->Name);

    // copy productions
    for(vector<NonTerminal*>::const_iterator i = cfg.nonterminals.begin(); i != cfg.nonterminals.end(); i++)
    {
        NonTerminal* n = static_cast<NonTerminal*>(symbols[(*i)->Name]);
        for(vector<Production*>::iterator p = (*i)->productions.begin(); p != (*i)->productions.end(); p++)
        {
            Production* np = new Production();
            for(vector<Symbol*>::iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
                np->produced.push_back(symbols[(*s)->Name]);
            n->AddOrDeleteProduction(np);
        }
    }

    // set start symbol
    startsymbol = static_cast<NonTerminal*>(symbols[cfg.startsymbol->Name]);

    return *this;
}






bool ContextFreeGrammars::ContextFreeGrammar::LoadFromStream(istream& is)
{
    Clear();

    string input;
    Symbol* symbol = NULL;
    NonTerminal* nonterminal = NULL;
    Production* production = NULL;

    for(is >> input; input != "%%"; is >> input)
        AddTerminal(input);

    reading_nonterminal:
        is >> input;
        if(input == "%%")
            goto finished;
        if(symbols.find(input) == symbols.end())
            AddNonTerminal(input);
        symbol = symbols[input];
        nonterminal = dynamic_cast<NonTerminal*>(symbol);
        if(nonterminal == NULL)
        {
            cerr << "Symbol " << input << " already defined as terminal." << endl;
            return false;
        }

        is >> input;
        if(input == ";") // broken / empty nonterminal
        {
            goto reading_nonterminal;
        }
        if(input != ":")
        {
            cerr << "expected : but got " << input << " after nonterminal " << input << endl;
            return false;
        }

    reading_new_production:
        production = new Production();

    reading_production_part:
        is >> input;
        if(input == ";")
        {
            nonterminal->AddOrDeleteProduction(production);
            goto reading_nonterminal;
        }
        if(input == "|")
        {
            nonterminal->AddOrDeleteProduction(production);
            goto reading_new_production;
        }

        if(symbols.find(input) == symbols.end())
            AddNonTerminal(input);
        symbol = symbols[input];

        production->produced.push_back(symbol);
        goto reading_production_part;

    finished:
        is >> input;
        symbol = symbols[input];
        nonterminal = dynamic_cast<NonTerminal*>(symbol);
        if(nonterminal == NULL)
        {
            cerr << "start-symbol is " << input << " but must be a non-terminal" << endl;
            return false;
        }
        this->startsymbol = nonterminal;


    sanity_check:
        for(vector<NonTerminal*>::iterator i = nonterminals.begin(); i != nonterminals.end(); i++)
            if((*i)->productions.size() <= 0)
                cerr << "Warning: Non-Terminal " << (*i)->Name << " has no productions. It was probably meant as terminal." << endl;
        return true;
}

void ContextFreeGrammars::ContextFreeGrammar::WriteToStream(ostream& os)
{
    for(vector<Terminal*>::iterator t = terminals.begin(); t != terminals.end(); t++)
        os << (*t)->Name << " ";
    os << endl << "%%" << endl;

    for(vector<NonTerminal*>::iterator n = nonterminals.begin(); n != nonterminals.end(); n++)
    {
        os << (*n)->Name;
        bool first = true;
        for(vector<Production*>::iterator p = (*n)->productions.begin(); p != (*n)->productions.end(); p++)
        {
            os << (first?" :":" |");
            for(vector<Symbol*>::iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
                os << " " << (*s)->Name;
            first = false;
        }
        os << " ;" << endl;
    }
    os << "%%" << endl << startsymbol->Name << endl;
}






ContextFreeGrammars::Terminal* ContextFreeGrammars::ContextFreeGrammar::AddTerminal(std::string Name)
{
    Terminal* t = new Terminal(Name);
    terminals.push_back(t);
    symbols[Name] = t;
    return t;
}
ContextFreeGrammars::NonTerminal* ContextFreeGrammars::ContextFreeGrammar::AddNonTerminal(std::string Name)
{
    NonTerminal* nt = new NonTerminal(Name);
    nonterminals.push_back(nt);
    symbols[Name] = nt;
    return nt;
}
void ContextFreeGrammars::ContextFreeGrammar::RemoveSymbol(Symbol* symbol)
{
    bool productionErased = false;
    for(vector<NonTerminal*>::iterator n = nonterminals.begin(); n != nonterminals.end(); n += !productionErased)
    {
        productionErased = false;
        for(vector<Production*>::iterator p = (*n)->productions.begin(); (!productionErased) && p != (*n)->productions.end(); p++)
        {
            for(vector<Symbol*>::iterator s = (*p)->produced.begin(); (!productionErased) && s != (*p)->produced.end(); s++)
                if(*s == symbol)
                {
                    delete *p;
                    (*n)->productions.erase(p);
                    productionErased = true; // this aborts the outer loop (necessary because the iterator is invalidated)
                }
        }
    }

    symbols.erase(symbol->Name);
    vector<NonTerminal*>::iterator ntm = find(nonterminals.begin(), nonterminals.end(), symbol);
    if(ntm != nonterminals.end())
        nonterminals.erase(ntm);
    else
        terminals.erase(find(terminals.begin(), terminals.end(), symbol));
    delete symbol;
}






vector<ContextFreeGrammars::NonTerminal*> ContextFreeGrammars::ContextFreeGrammar::FindUnproductiveNonTerminals() const
{
    map<Symbol*, bool> productive;

    // all Terminals are "productive"
    for(vector<Terminal*>::const_iterator t = terminals.begin(); t != terminals.end(); t++)
        productive[*t] = true;
    // non-terminals are unproductive until proven otherwise
    for(vector<NonTerminal*>::const_iterator n = nonterminals.begin(); n != nonterminals.end(); n++)
        productive[*n] = false;

    bool found_new = false;
    do
    {
        found_new = false;

        for(vector<NonTerminal*>::const_iterator n = nonterminals.begin(); n != nonterminals.end(); n++)
        {
            if(productive[*n]) // nonterminal is already marked as productive
                continue;

            for(vector<Production*>::iterator p = (*n)->productions.begin(); p != (*n)->productions.end(); p++)
            {
                bool productionproductive = true;
                for(vector<Symbol*>::iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
                    if(!productive[*s])
                    {
                        productionproductive = false;
                        break;
                    }
                if(productionproductive)
                {
                    productive[*n] = true;
                    found_new = true;
                    break;
                }
            }
        }
    } while(found_new);


    vector<NonTerminal*> result;
    for(vector<NonTerminal*>::const_iterator n = nonterminals.begin(); n != nonterminals.end(); n++)
        if(!productive[*n])
            result.push_back(*n);
    return result;
}
void ContextFreeGrammars::ContextFreeGrammar::RemoveUnproductiveNonTerminals()
{
    vector<NonTerminal*> Unproductive = FindUnproductiveNonTerminals();
    for(vector<NonTerminal*>::iterator n = Unproductive.begin(); n != Unproductive.end(); n++)
        RemoveSymbol(*n);
}






vector<ContextFreeGrammars::Symbol*> ContextFreeGrammars::ContextFreeGrammar::FindUnreachableSymbols() const
{
    map<Symbol*, bool> reachable;
    for(map<string,Symbol*>::const_iterator i = symbols.begin(); i != symbols.end(); i++)
        reachable[i->second] = false;
    reachable[startsymbol] = true;

    queue<Symbol*> queue;
    for(queue.push(startsymbol); !queue.empty(); queue.pop())
    {
        Symbol* front = queue.front();
        NonTerminal* nFront = dynamic_cast<NonTerminal*>(front);
        if(nFront == NULL) // Terminal
            continue;

        for(vector<Production*>::iterator p = nFront->productions.begin(); p != nFront->productions.end(); p++)
        {
            for(vector<Symbol*>::iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
            {
                if(reachable[*s]) // already marked
                    continue;
                reachable[*s] = true;
                queue.push(*s);
            }
        }
    }

    vector<Symbol*> result;
    for(map<Symbol*,bool>::iterator i = reachable.begin(); i != reachable.end(); i++)
        if(!reachable[i->first])
            result.push_back(i->first);
    return result;
}
void ContextFreeGrammars::ContextFreeGrammar::RemoveUnreachableSymbols()
{
    vector<Symbol*> Unreachable = FindUnreachableSymbols();
    for(vector<Symbol*>::iterator s = Unreachable.begin(); s != Unreachable.end(); s++)
        RemoveSymbol(*s);
}






set<const ContextFreeGrammars::NonTerminal*> ContextFreeGrammars::ContextFreeGrammar::FindEpsilonProducers() const
{
    set<const NonTerminal*> result;

    // collect non-terminals A that have explicit  A -> epsilon  productions
    for(vector<NonTerminal*>::const_iterator i = nonterminals.begin(); i != nonterminals.end(); i++)
        for(vector<Production*>::const_iterator p = (*i)->productions.begin(); p != (*i)->productions.end(); p++)
            if((*p)->produced.size() <= 0)
                result.insert(*i);

    bool foundNew = result.size() > 0;
    while(foundNew)
    {
        foundNew = false;
        for(vector<NonTerminal*>::const_iterator i = nonterminals.begin(); !foundNew && i != nonterminals.end(); i++)
        {
            if(result.find(*i) != result.end()) // already recognized as epsilon-producer
                continue;

            for(vector<Production*>::const_iterator p = (*i)->productions.begin(); !foundNew && p != (*i)->productions.end(); p++)
            {
                bool producesEpsilon = true;
                for(vector<Symbol*>::const_iterator s = (*p)->produced.begin(); producesEpsilon && s != (*p)->produced.end(); s++)
                    if(result.find(dynamic_cast<NonTerminal*>(*s)) == result.end())
                        producesEpsilon = false; // part of the production is not (yet) an epsilon-producer

                if(producesEpsilon)
                {
                    result.insert(*i);
                    foundNew = true;
                }
            }
        }
    }

    return result;
}
void ContextFreeGrammars::ContextFreeGrammar::AddEpsilonCleanedProductions(ContextFreeGrammars::Production* prod, int position,
                             vector<ContextFreeGrammars::Symbol*>& stack, set<const ContextFreeGrammars::NonTerminal*>& epsProducers,
                             NonTerminal* nt, map<Symbol*, Symbol*>& correspondingsymbol)
{
    if(position >= prod->produced.size())
    {
        if(stack.size() > 0) // otherwise this stack corresponds to epsilon-production => eliminated
        {
            Production* newProd = new Production();
            newProd->produced = stack;
            nt->AddOrDeleteProduction(newProd);
        }
        return;
    }

    NonTerminal* prodsymnt = dynamic_cast<NonTerminal*>(prod->produced[position]);

    stack.push_back(correspondingsymbol[prod->produced[position]]);
    AddEpsilonCleanedProductions(prod, position+1, stack, epsProducers, nt, correspondingsymbol);
    stack.pop_back();

    if(prodsymnt != NULL && epsProducers.find(prodsymnt) != epsProducers.end())
        AddEpsilonCleanedProductions(prod, position+1, stack, epsProducers, nt, correspondingsymbol);
}
ContextFreeGrammars::ContextFreeGrammar ContextFreeGrammars::ContextFreeGrammar::EpsilonElimination() const
{
    set<const ContextFreeGrammars::NonTerminal*> epsProducers = FindEpsilonProducers();
    if(epsProducers.size() <= 0)
        return *this;
    ContextFreeGrammar result;
    map<Symbol*, Symbol*> correspondingsymbol;

    // copy terminals
    for(vector<Terminal*>::const_iterator i = terminals.begin(); i != terminals.end(); i++)
        correspondingsymbol[*i] = result.AddTerminal((*i)->Name);

    // copy nonterminals
    for(vector<NonTerminal*>::const_iterator i = nonterminals.begin(); i != nonterminals.end(); i++)
        correspondingsymbol[*i] = result.AddNonTerminal((*i)->Name);

    for(vector<NonTerminal*>::const_iterator i = nonterminals.begin(); i != nonterminals.end(); i++)
    {
        NonTerminal* nt = static_cast<NonTerminal*>(correspondingsymbol[*i]);
        vector<Symbol*> stack;
        for(vector<Production*>::const_iterator p = (*i)->productions.begin(); p != (*i)->productions.end(); p++)
            result.AddEpsilonCleanedProductions(*p, 0, stack, epsProducers, nt, correspondingsymbol);
    }

    if(epsProducers.find(startsymbol) != epsProducers.end())
    {
        NonTerminal *s0 = result.AddNonTerminal(startsymbol->Name + "_0");

        Production *s0sProduction = new Production();
        s0sProduction->produced.push_back(correspondingsymbol[startsymbol]);
        s0->AddOrDeleteProduction(s0sProduction);

        Production *s0epsProduction = new Production();
        s0->AddOrDeleteProduction(s0epsProduction);

        result.startsymbol = s0;
    }
    else
    {
        result.startsymbol = static_cast<NonTerminal*>(correspondingsymbol[startsymbol]);
    }

    return result;
}






ContextFreeGrammars::ContextFreeGrammar ContextFreeGrammars::ContextFreeGrammar::UnitProductionElimination() const
{
    ContextFreeGrammar result;
    map<Symbol*, Symbol*> correspondingsymbol;
    map<Symbol*, set<NonTerminal*> > reachable;

    // copy terminals
    for(vector<Terminal*>::const_iterator i = terminals.begin(); i != terminals.end(); i++)
        correspondingsymbol[*i] = result.AddTerminal((*i)->Name);

    // copy nonterminals
    for(vector<NonTerminal*>::const_iterator i = nonterminals.begin(); i != nonterminals.end(); i++)
        correspondingsymbol[*i] = result.AddNonTerminal((*i)->Name);

    for(vector<NonTerminal*>::const_iterator i = nonterminals.begin(); i != nonterminals.end(); i++)
    {
        reachable[*i].insert(*i);
        queue<Symbol*> queue;
        for(queue.push(*i); !queue.empty(); queue.pop())
        {
            NonTerminal* front = dynamic_cast<NonTerminal*>(queue.front());

            for(vector<Production*>::iterator p = front->productions.begin(); p != front->productions.end(); p++)
            {
                if((*p)->produced.size() != 1) // not a unit production
                    continue;
                Symbol *prod = *((*p)->produced.begin());
                NonTerminal *nprod = dynamic_cast<NonTerminal*>(prod);
                if(nprod == NULL) // not a non-terminal
                    continue;
                if(reachable[*i].find(nprod) != reachable[*i].end()) // already marked
                    continue;
                reachable[*i].insert(nprod);
                queue.push(nprod);
            }
        }
    }

    // copy (non-unit) productions
    for(vector<NonTerminal*>::const_iterator i = nonterminals.begin(); i != nonterminals.end(); i++)
    {
        for(set<NonTerminal*>::const_iterator j = reachable[*i].begin(); j != reachable[*i].end(); j++)
        {
            for(vector<Production*>::const_iterator p = (*j)->productions.begin(); p != (*j)->productions.end(); p++)
            {
                if((*p)->produced.size() == 1) // unit productions are not copied
                    if( dynamic_cast<NonTerminal*>(*((*p)->produced.begin())) != NULL )
                        continue;

                Production *prodCopy = new Production();
                for(vector<Symbol*>::const_iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
                    prodCopy->produced.push_back(correspondingsymbol[*s]);
                NonTerminal* corresponding = static_cast<NonTerminal*>(correspondingsymbol[*i]);
                corresponding->AddOrDeleteProduction(prodCopy);
            }
        }
    }

    result.startsymbol = static_cast<NonTerminal*>(correspondingsymbol[startsymbol]);
    return result;
}






void ContextFreeGrammars::ContextFreeGrammar::Produce(int depth, const vector<const Symbol*>& stack, void callback(const vector<const Symbol*>&)) const
{
    if(depth < 0)
        return;

    bool isTerminalWord = true;
    vector<const Symbol*> newstack;
    for(vector<const Symbol*>::const_iterator i = stack.begin(); i != stack.end(); i++)
    {
        const NonTerminal* n = dynamic_cast<const NonTerminal*>(*i);
        if(n == NULL) // terminals are not
            continue;
        isTerminalWord = false;

        for(vector<Production*>::const_iterator p = n->productions.begin(); p != n->productions.end(); p++)
        {
            newstack.clear();
            for(vector<const Symbol*>::const_iterator k = stack.begin(); k != i; k++) // front of old stack
                newstack.push_back(*k);
            for(vector<Symbol*>::const_iterator k = (*p)->produced.begin(); k != (*p)->produced.end(); k++) // replacement from current production
                newstack.push_back(*k);
            for(vector<const Symbol*>::const_iterator k = i+1; k != stack.end(); k++) // tail of old stack
                newstack.push_back(*k);
            Produce(depth-1, newstack, callback);
        }
    }

    if(isTerminalWord)
        callback(stack);
}
void ContextFreeGrammars::ContextFreeGrammar::Produce(int depth, void callback(const vector<const ContextFreeGrammars::Symbol*>&)) const
{
    vector<const Symbol*> stack(1, startsymbol);
    Produce(depth, stack, callback);
}






ContextFreeGrammars::ContextFreeGrammar ContextFreeGrammars::ContextFreeGrammar::ChomskyNormalForm() const
{
    const ContextFreeGrammar *temp0 = this;
/*
    temp0 = temp0.EpsilonElimination();
    temp0 = temp0.UnitProductionElimination();
*/
    ContextFreeGrammar temp1;
    map<Symbol*, NonTerminal*> correspondingsymbol;

    // copy nonterminals
    for(vector<NonTerminal*>::const_iterator i = temp0->nonterminals.begin(); i != temp0->nonterminals.end(); i++)
        correspondingsymbol[*i] = temp1.AddNonTerminal((*i)->Name);

    // create R_a -> a  productions
    for(vector<Terminal*>::const_iterator i = temp0->terminals.begin(); i != temp0->terminals.end(); i++)
    {
        Terminal* t = temp1.AddTerminal((*i)->Name);
        NonTerminal* nt = temp1.AddNonTerminal("R_{" + (*i)->Name + "}");
        Production* p = new Production();
        p->produced.push_back(t);
        nt->AddOrDeleteProduction(p);
        correspondingsymbol[*i] = nt;
    }

    // break overlong productions apart
    for(vector<NonTerminal*>::const_iterator i = temp0->nonterminals.begin(); i != temp0->nonterminals.end(); i++)
    {
        NonTerminal* nt = correspondingsymbol[*i];
        int pidx = 1;

        for(vector<Production*>::iterator p = (*i)->productions.begin(); p != (*i)->productions.end(); p++, pidx++)
        {
            if((*p)->produced.size() >= 2)
            {
                // A -> X1 X2 X3 X4 becomes  A -> X1 A1
                //                          A1 -> X2 A2
                //                          A2 -> X3 X4

                string prefix = nt->Name + "_{" + std::to_string(pidx) + ",";
                NonTerminal *lastnonterminal = nt;

                for(int j = 1; j < (*p)->produced.size()-1; ++j)
                {
                    // create next nonterminal Aj
                    NonTerminal *nj = temp1.AddNonTerminal(prefix + std::to_string(j) + "}");

                    // create the production lastnonterminal -> Xj Aj
                    Production *pj = new Production();
                    pj->produced.push_back(correspondingsymbol[(*p)->produced[j-1]]);
                    pj->produced.push_back(nj);

                    // add the production to lastnonterminal
                    lastnonterminal->AddOrDeleteProduction(pj);
                    lastnonterminal = nj;
                }

                // create An-2 -> Xn-1 Xn
                int last = (*p)->produced.size()-1;
                Production *plast = new Production();
                plast->produced.push_back(correspondingsymbol[(*p)->produced[last-1]]);
                plast->produced.push_back(correspondingsymbol[(*p)->produced[last]]);
                lastnonterminal->AddOrDeleteProduction(plast);
            }
            else
            {
                Production *prodCopy = new Production();
                for(vector<Symbol*>::iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
                    prodCopy->produced.push_back(correspondingsymbol[*s]);
                nt->AddOrDeleteProduction(prodCopy);
            }
        }
    }

    temp1.startsymbol = correspondingsymbol[temp0->startsymbol];

    ContextFreeGrammar temp2 = temp1.EpsilonElimination();
    ContextFreeGrammar temp3 = temp2.UnitProductionElimination();
    temp3.RemoveUnproductiveNonTerminals();
    temp3.RemoveUnreachableSymbols();
    return temp3;
}




bool ContextFreeGrammars::ContextFreeGrammar::Producable(const vector<string>& w) const
{
    vector<Symbol*> s;
    for(vector<string>::const_iterator i = w.begin(); i != w.end(); i++)
    {
        map<string, Symbol*>::const_iterator it = symbols.find(*i);
        Symbol* is = it->second;
        s.push_back(is);
    }
    return Producable(s);
}

bool ContextFreeGrammars::ContextFreeGrammar::Producable(const vector<Symbol*>& w) const
{
    ContextFreeGrammar cnf = ChomskyNormalForm();

    vector<vector<set<NonTerminal*> > > Table;


}



