
#include "LL1ParserGenerator.h"
using namespace std;

ContextFreeGrammars::LL1ParserGenerator::LL1ParserGenerator(const ContextFreeGrammar& forcfg)
    : cfg(forcfg)
{
    for(vector<NonTerminal*>::iterator nt = cfg.nonterminals.begin(); nt != cfg.nonterminals.end(); nt++)
        CollectNonTerminalFirst1(*nt);

    for(vector<NonTerminal*>::iterator nt = cfg.nonterminals.begin(); nt != cfg.nonterminals.end(); nt++)
        CollectNonTerminalFollow1(*nt);
}

ContextFreeGrammars::LL1ParserGenerator::~LL1ParserGenerator()
{
}






void ContextFreeGrammars::LL1ParserGenerator::CollectNonTerminalFollow1(const NonTerminal *targetnt)
{
    set<const NonTerminal*> VisitedNonterminals;

    VisitedNonterminals.insert(targetnt);
    queue<const NonTerminal*> ntqueue;
    for(ntqueue.push(targetnt); !ntqueue.empty(); ntqueue.pop())
    {
        const NonTerminal* front = ntqueue.front();

        for(vector<NonTerminal*>::const_iterator n = cfg.nonterminals.begin(); n != cfg.nonterminals.end(); n++)
        {
            for(vector<Production*>::const_iterator p = (*n)->productions.begin(); p != (*n)->productions.end(); p++)
            {
                for(vector<Symbol*>::const_iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
                {
                    if(*s != front)
                        continue;

                    bool EpsilonFollows = true;
                    vector<Symbol*>::const_iterator sfollow = s;
                    for(sfollow++ ; EpsilonFollows && sfollow != (*p)->produced.end(); sfollow++)
                    {
                        // insert the terminal
                        Terminal *t = dynamic_cast<Terminal*>(*sfollow);
                        if(t != NULL)
                        {
                            nonterminalfollow1[targetnt].insert(t);
                            EpsilonFollows = false;
                            break;
                        }

                        // insert all elements of the first1 set of the nonterminal
                        NonTerminal *nt = static_cast<NonTerminal*>(*sfollow);
                        map<const NonTerminal*, set<Terminal*> >::const_iterator sourceset = nonterminalfirst1.find(nt);
                        bool ContainsEpsilon = false;
                        for(set<Terminal*>::const_iterator f = sourceset->second.begin(); f != sourceset->second.end(); f++)
                            if(*f != NULL)
                                nonterminalfollow1[targetnt].insert(*f);
                            else
                                ContainsEpsilon = true;

                        EpsilonFollows = ContainsEpsilon;
                    }

                    if(EpsilonFollows && VisitedNonterminals.find(*n) == VisitedNonterminals.end())
                    {
                        ntqueue.push(*n);
                        VisitedNonterminals.insert(*n);

                        if(*n == cfg.startsymbol) // the symbol can be the last one
                            nonterminalfollow1[targetnt].insert(NULL);
                    }
                }
            }
        }

    }
}

void ContextFreeGrammars::LL1ParserGenerator::CollectNonTerminalFirst1(const NonTerminal *targetnt)
{
    set<const NonTerminal*> VisitedNonterminals;
    set<const NonTerminal*> epsProducers = cfg.FindEpsilonProducers();

    VisitedNonterminals.insert(targetnt);
    queue<const NonTerminal*> ntqueue;
    for(ntqueue.push(targetnt); !ntqueue.empty(); ntqueue.pop())
    {
        const NonTerminal* front = ntqueue.front();
        for(vector<Production*>::const_iterator p = front->productions.begin(); p != front->productions.end(); p++)
        {
            for(vector<Symbol*>::const_iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
            {
                Terminal *t = dynamic_cast<Terminal*>(*s);
                if(t != NULL)
                {
                    nonterminalfirst1[targetnt].insert(t);
                    break;
                }

                NonTerminal *nt = static_cast<NonTerminal*>(*s);
                if(VisitedNonterminals.find(nt) == VisitedNonterminals.end())
                {
                    VisitedNonterminals.insert(nt);
                    ntqueue.push(nt);
                }
                if(epsProducers.find(nt) == epsProducers.end())
                    break;
            }
        }
    }

    if(epsProducers.find(targetnt) != epsProducers.end())
        nonterminalfirst1[targetnt].insert(NULL);
}

set<ContextFreeGrammars::Terminal*> ContextFreeGrammars::LL1ParserGenerator::first1(const vector<Symbol*>& w) const
{
    set<Terminal*> result;

    bool EpsilonProducible = true;
    for(vector<Symbol*>::const_iterator i = w.begin(); EpsilonProducible && i != w.end(); i++)
    {
        Terminal *t = dynamic_cast<Terminal*>(*i);
        if(t != NULL)
        {
            result.insert(t);
            EpsilonProducible = false;
            break;
        }

        NonTerminal *nt = static_cast<NonTerminal*>(*i);
        bool ContainsEpsilon = false;

        map<const NonTerminal*, set<Terminal*> >::const_iterator sourceset = nonterminalfirst1.find(nt);
        for(set<Terminal*>::const_iterator f = sourceset->second.begin(); f != sourceset->second.end(); f++)
            if(*f != NULL)
                result.insert(*f);
            else
                ContainsEpsilon = true;

        if(!ContainsEpsilon)
            EpsilonProducible = false;
    }

    if(EpsilonProducible)
        result.insert(NULL);
    return result;
}

set<ContextFreeGrammars::Terminal*> ContextFreeGrammars::LL1ParserGenerator::lookahead1(const ContextFreeGrammars::NonTerminal* A, const vector<ContextFreeGrammars::Symbol*>& beta) const
{
    set<Terminal*> result = first1(beta);
    if(result.find(NULL) != result.end())
    {
        result.erase(NULL);
        map<const NonTerminal*, set<Terminal*> >::const_iterator sourceset = nonterminalfollow1.find(A);
        for(set<Terminal*>::iterator f = sourceset->second.begin(); f != sourceset->second.end(); f++)
            result.insert(*f);
    }
    return result;
}





bool ContextFreeGrammars::LL1ParserGenerator::IsLL1Grammar() const
{
    for(vector<NonTerminal*>::const_iterator nt = cfg.nonterminals.begin(); nt != cfg.nonterminals.end(); nt++)
    {
        for(vector<Production*>::const_iterator p1 = (*nt)->productions.begin(); p1 != (*nt)->productions.end(); p1++)
        {
            vector<Production*>::const_iterator p2 = p1;
            for(p2++; p2 != (*nt)->productions.end(); p2++)
            {
                set<Terminal*> p1lookahead = lookahead1(*nt, (*p1)->produced);
                set<Terminal*> p2lookahead = lookahead1(*nt, (*p2)->produced);

                for(set<Terminal*>::iterator i = p1lookahead.begin(); i != p1lookahead.end(); i++)
                    if(p2lookahead.find(*i) != p2lookahead.end()) // *i is in the intersection of the lookahead-sets => not an LL(1) grammar
                        return false;
            }
        }
    }
    return true;
}




void ContextFreeGrammars::LL1ParserGenerator::GenerateRecursiveDescentParser(std::ostream& os) const
{
    os << "// Generated by LL1ParserGenerator by Viktor \"AlgorithMan\" Engelmann" << endl;
    os << "//                                 http://www.AlgorithMan.de" << endl;
    os << "#include<iostream>" << endl << endl;

    os << "bool verbose = false;" << endl << endl;

    os << "std::string ConsumeToken(std::string* lookahead, std::istream& is, int depth)" << endl;
    os << "{" << endl;
    os << "    if(verbose)"<< endl;
    os << "        std::cout << std::string(2*depth, ' ') << *lookahead << std::endl;" << endl << endl;

    os << "    std::string result = *lookahead;" << endl;
    os << "    if(is.eof())" << endl;
    os << "        *lookahead = \"\";" << endl;
    os << "    else" << endl;
    os << "    {" << endl;
    os << "        is >> *lookahead;" << endl;
    os << "        is.peek();" << endl;
    os << "        if(is.fail())" << endl;
    os << "            *lookahead = \"\";" << endl;
    os << "    }" << endl << endl;

    os << "    return result;" << endl;
    os << "}" << endl << endl;


    // forward declarations
    os << "// forward declarations" << endl;
    for(vector<NonTerminal*>::const_iterator n = cfg.nonterminals.begin(); n != cfg.nonterminals.end(); n++)
        os << "bool Parse" << (*n)->Name << "(std::string* lookahead, std::istream& is, int depth);" << endl;
    os << endl;


    for(vector<NonTerminal*>::const_iterator n = cfg.nonterminals.begin(); n != cfg.nonterminals.end(); n++)
    {
        os << "bool Parse" << (*n)->Name << "(std::string* lookahead, std::istream& is, int depth)" << endl << "{" << endl;
        os << "    if(verbose)" << endl;
        os << "        std::cout << std::string(2*depth, ' ') << \"" << (*n)->Name << ", look-ahead=\\\"\" << *lookahead << \"\\\"\" << std::endl;" << endl;

        bool firstproduction = true;
        for(vector<Production*>::const_iterator p = (*n)->productions.begin(); p != (*n)->productions.end(); p++)
        {
            os << endl << "    // " << (*n)->Name << " :";
            for(vector<Symbol*>::const_iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
                os << " " << (*s)->Name;

            os << endl << "    " << (firstproduction ? "" : "else ") << "if(";
            firstproduction = false;

            set<Terminal*> plookahead = lookahead1(*n, (*p)->produced);
            bool firstlookahead = true;
            for(set<Terminal*>::iterator la = plookahead.begin(); la != plookahead.end(); la++)
                if(*la != NULL)
                {
                    os << (firstlookahead ? "" : " || " ) << "*lookahead == \"" << (*la)->Name << "\"";
                    firstlookahead = false;
                }
            if(plookahead.find(NULL) != plookahead.end())
                os << (firstlookahead ? "" : " || " ) << "*lookahead == \"\")" << endl;
            else
                os << ")" << endl;

            os << "    {" << endl;

            for(vector<Symbol*>::const_iterator s = (*p)->produced.begin(); s != (*p)->produced.end(); s++)
            {
                Terminal* t = dynamic_cast<Terminal*>(*s);
                if(t != NULL)
                    os << "        if(ConsumeToken(lookahead, is, depth+1) != \"" << t->Name << "\") return false;" << endl;
                else
                {
                    NonTerminal* nt = static_cast<NonTerminal*>(*s);
                    os << "        if(!Parse" << nt->Name << "(lookahead, is, depth+1)) return false;" << endl;
                }
            }
            os << "        return true;" << endl;
            os << "    }" << endl;
        }

        os << endl << "    return false;" << endl;
        os << "}" << endl << endl;
    }


    os << "int main(int argc, char** argv)" << endl;
    os << "{" << endl;
    os << "    std::cout << \"LL(1) Parser generated by LL1ParserGenerator by Viktor \\\"AlgorithMan\\\" Engelmann\" << std::endl;" << endl;
    os << "    std::cout << \"                                             http://www.AlgorithMan.de\" << std::endl;" << endl;
    os << "    std::string lookahead;" << endl;
    os << "    std::cin >> lookahead;" << endl;
    os << "    if(argc > 1 && std::string(argv[1]) == \"-v\")" << endl;
    os << "        verbose = true;" << endl;
    os << "    std::cout << \"Parsing stdin...\\n\";" << endl;
    os << "    bool result = Parse" << cfg.startsymbol->Name << "(&lookahead, std::cin, 0);" << endl;
    os << "    std::cout << (result ? \"success\" : \"failed\") << std::endl;" << endl;
    os << "    return result ? 0 : 1;" << endl;
    os << "}" << endl << endl;
}











